﻿using System;
using System.Windows.Forms;
using System.IO;
using System.Net;
using System.Security.Cryptography;

namespace NUS_Downloader
{
    public partial class Form1 : Form
    {
        const string NUSURL = "http://nus.cdn.shop.wii.com/ccs/download/";
        const string DSiNUSURL = "http://nus.cdn.t.shop.nintendowifi.net/ccs/download/";
        string version = "v1.1a";
        WebClient generalWC = new WebClient();
        static RijndaelManaged rijndaelCipher;
        static bool dsidecrypt = false;

        public struct WADHeader
        {
            public int HeaderSize;
            public int WadType;
            public int CertChainSize;
            public int Reserved;
            public int TicketSize;
            public int TMDSize;
            public int DataSize;
            public int FooterSize;
        };

        const string certs_MD5 = "7677AD47BAA5D6E3E313E72661FBDC16";

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            this.Text = "NUSD - " + version + " - WB3000";

            // Directory stuff
            string currentdir = Application.StartupPath;
            if (currentdir.EndsWith(Convert.ToString(Path.DirectorySeparatorChar)) == false)
                currentdir += Path.DirectorySeparatorChar;

            if (File.Exists(currentdir + "cert.sys") == false)
            {
                foreach (Control ctrl in this.Controls)
                {
                    ctrl.Enabled = false;
                }
                getcerts.Enabled = true;
                WriteStatus("You do not have a certs file. Press the button below to generate a cert file!");
            }
            else if (verifyMd5Hash(currentdir + "cert.sys", certs_MD5) == false)
            {
                foreach (Control ctrl in this.Controls)
                {
                    ctrl.Enabled = false;
                }
                getcerts.Enabled = true;
                WriteStatus("Your certs file is corrupted/invalid. Press the button below to generate a cert file!");
            }
            else
            {
                getcerts.Visible = false;
                WriteStatus("Certs file is present and intact.");
            }

            if (File.Exists(currentdir + "key.bin") == false)
            {
                WriteStatus("Common Key (key.bin) missing! Decryption disabled!");
                decryptbox.Visible = false;
            }
            else
            {
                WriteStatus("Common Key detected.");
            }

            if (File.Exists(currentdir + "dskey.bin") == false)
            {
                //WriteStatus("Common Key (key.bin) missing! Decryption disabled!");
                //decryptbox.Visible = false;
            }
            else
            {
                WriteStatus("DSi Common Key detected.");
                dsidecrypt = true;
            }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            OpenFileDialog opentmd = new OpenFileDialog();
            opentmd.Filter = "TMD Files|tmd";
            opentmd.Title = "Open TMD";
            if (opentmd.ShowDialog() != DialogResult.Cancel)
            {
                // Read the tmd as a stream...
                FileStream fs = File.OpenRead(opentmd.FileName);
                byte[] tmd = ReadFully(fs, 20);
                WriteStatus("TMD Loaded (" + tmd.Length + " bytes)");

                // Read ID...
                titleidbox.Text = "";
                for (int x = 396; x < 404; x++)
                {
                    titleidbox.Text += MakeProperLength(ConvertToHex(Convert.ToString(tmd[x])));
                }
                WriteStatus("Title ID: " + titleidbox.Text);
                ReadIDType(titleidbox.Text);

                // Read Title Version...
                string tmdversion = "";
                for (int x = 476; x < 478; x++)
                {
                    tmdversion += MakeProperLength(ConvertToHex(Convert.ToString(tmd[x])));
                }
                titleversion.Text = Convert.ToString(int.Parse(tmdversion, System.Globalization.NumberStyles.HexNumber));

                // Read System Version (Needed IOS)
                string sysversion = "";
                for (int i = 0; i < 8; i++)
                {
                    sysversion += MakeProperLength(ConvertToHex(Convert.ToString(tmd[0x184 + i])));
                }
                sysversion = Convert.ToString(int.Parse(sysversion.Substring(14, 2), System.Globalization.NumberStyles.HexNumber));
                if ((sysversion != "0") || (sysversion != "0"))
                    WriteStatus("Requires: IOS" + sysversion);

                // Read Content #...
                string contentstrnum = "";
                for (int x = 478; x < 480; x++)
                {
                    //contentstrnum += MakeProperLength(ConvertToHex(Convert.ToString(tmd[x])));
                    contentstrnum += TrimLeadingZeros(Convert.ToString(tmd[x]));
                }
                WriteStatus("Content Count: " + contentstrnum);

                string[] tmdcontents = GetContentNames(tmd, Convert.ToInt32(contentstrnum));
                string[] tmdsizes = GetContentSizes(tmd, Convert.ToInt32(contentstrnum));
                byte[] tmdhashes = GetContentHashes(tmd, Convert.ToInt32(contentstrnum));
                byte[] tmdindices = GetContentIndices(tmd, Convert.ToInt32(contentstrnum));

                for (int i = 0; i < Convert.ToInt32(contentstrnum); i++)
                {
                    WriteStatus("   Content " + (i + 1) + ": " + tmdcontents[i] + " (" + Convert.ToString(int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber)) + " bytes)");
                    byte[] hash = new byte[20];
                    for (int x = 0; x < 20; x++)
                    {
                        hash[x] = tmdhashes[(i*20)+x];
                    }
                    WriteStatus("  - Hash: " + DisplayBytes(hash));
                    WriteStatus("  - Index: " + tmdindices[i]);

                }
            }
        }

        private void WriteStatus(string Update)
        {
            if (statusbox.Text == "")
                statusbox.Text = Update;
            else
                statusbox.Text += "\r\n" + Update;

            // Scroll to end of text box.
            statusbox.SelectionStart = statusbox.TextLength;
            statusbox.ScrollToCaret();
        }

        /// <summary>
        /// Reads data from a stream until the end is reached. The
        /// data is returned as a byte array. An IOException is
        /// thrown if any of the underlying IO calls fail.
        /// </summary>
        /// <param name="stream">The stream to read data from</param>
        /// <param name="initialLength">The initial buffer length</param>
        public static byte[] ReadFully(Stream stream, int initialLength)
        {
            // If we've been passed an unhelpful initial length, just
            // use 32K.
            if (initialLength < 1)
            {
                initialLength = 32768;
            }

            byte[] buffer = new byte[initialLength];
            int read = 0;

            int chunk;
            while ((chunk = stream.Read(buffer, read, buffer.Length - read)) > 0)
            {
                read += chunk;

                // If we've reached the end of our buffer, check to see if there's
                // any more information
                if (read == buffer.Length)
                {
                    int nextByte = stream.ReadByte();

                    // End of stream? If so, we're done
                    if (nextByte == -1)
                    {
                        return buffer;
                    }

                    // Nope. Resize the buffer, put in the byte we've just
                    // read, and continue
                    byte[] newBuffer = new byte[buffer.Length * 2];
                    Array.Copy(buffer, newBuffer, buffer.Length);
                    newBuffer[read] = (byte)nextByte;
                    buffer = newBuffer;
                    read++;
                }
            }
            // Buffer is now too big. Shrink it.
            byte[] ret = new byte[read];
            Array.Copy(buffer, ret, read);
            return ret;
        }

        private string MakeProperLength(string hex)
        {
            if (hex.Length == 1)
                hex = "0" + hex;

            return hex;
        }

        private string ConvertToHex(string decval)
        {
            //Console.Write("Decimal Value: " + decval);

            // Convert text string to unsigned integer
            int uiDecimal = System.Convert.ToInt32(decval);

            string hexval;
            // Format unsigned integer value to hex and show in another textbox
            //Console.Write("Hex Value: " + String.Format("{0:x2}", uiDecimal));
            hexval = String.Format("{0:x2}", uiDecimal);
            return hexval;
        }

        private void ReadIDType(string ttlid)
        {
            /*  Wiibrew TitleID Info...
                # 3 00000001: Essential system titles
                # 4 00010000 and 00010004 : Disc-based games
                # 5 00010001: Downloaded channels

                    * 5.1 000010001-Cxxx : Commodore 64 Games
                    * 5.2 000010001-Exxx : NeoGeo Games
                    * 5.3 000010001-Fxxx : NES Games
                    * 5.4 000010001-Hxxx : Channels
                    * 5.5 000010001-Jxxx : SNES Games
                    * 5.6 000010001-Nxxx : Nintendo 64 Games
                    * 5.7 000010001-Wxxx : WiiWare

                # 6 00010002: System channels
                # 7 00010004: Game channels and games that use them
                # 8 00010005: Downloaded Game Content
                # 9 00010008: "Hidden" channels
             */

            if (ttlid.Substring(0, 8) == "00000001")
            {
                WriteStatus("ID Type: System Title. BE CAREFUL!");
            }
            else if ((ttlid.Substring(0, 8) == "00010000") || (ttlid.Substring(0, 8) == "00010004"))
            {
                WriteStatus("ID Type: Disc-Based Game. Unlikely NUS Content!");
            }
            else if (ttlid.Substring(0, 8) == "00010001")
            {
                WriteStatus("ID Type: Downloaded Channel. Possible NUS Content.");
            }
            else if (ttlid.Substring(0, 8) == "00010002")
            {
                WriteStatus("ID Type: System Channel. BE CAREFUL!");
            }
            else if (ttlid.Substring(0, 8) == "00010004")
            {
                WriteStatus("ID Type: Game Channel. Unlikely NUS Content!");
            }
            else if (ttlid.Substring(0, 8) == "00010005")
            {
                WriteStatus("ID Type: Downloaded Game Content. Unlikely NUS Content!");
            }
            else if (ttlid.Substring(0, 8) == "00010008")
            {
                WriteStatus("ID Type: 'Hidden' Channel. Unlikely NUS Content!");
            }
            else
            {
                WriteStatus("ID Type: Unknown. Unlikely NUS Content!");
            }
        }

        private string TrimLeadingZeros(string num)
        {
            int startindex = 0;
            for (int i = 0; i < num.Length; i++)
            {
                if ((num[i] == 0) || (num[i] == '0'))
                    startindex += 1;
                else
                    break;
            }

            return num.Substring(startindex, (num.Length - startindex));
        }

        private string[] GetContentNames(byte[] tmdfile, int length)
        {
            string[] contentnames = new string[length];
            int startoffset = 484;

            for (int i = 0; i < length; i++)
            {
                contentnames[i] = MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset]))) + 
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 1]))) + 
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 2]))) + 
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 3])));
                startoffset += 36;
            }

            return contentnames;
        }

        private string[] GetContentSizes(byte[] tmdfile, int length)
        {
            string[] contentsizes = new string[length];
            int startoffset = 492;

            for (int i = 0; i < length; i++)
            {
                contentsizes[i] = MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 1]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 2]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 3]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 4]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 5]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 6]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 7])));
                contentsizes[i] = TrimLeadingZeros(contentsizes[i]);
                /*contentsizes[i] = Convert.ToString(tmdfile[startoffset]) +
                    Convert.ToString(tmdfile[startoffset + 1]) +
                    Convert.ToString(tmdfile[startoffset + 2]) +
                    Convert.ToString(tmdfile[startoffset + 3]) +
                    Convert.ToString(tmdfile[startoffset + 4]) +
                    Convert.ToString(tmdfile[startoffset + 5]) +
                    Convert.ToString(tmdfile[startoffset + 6]) +
                    Convert.ToString(tmdfile[startoffset + 7]);
                contentsizes[i] = TrimLeadingZeros(contentsizes[i]);  */
                startoffset += 36;
            }

            return contentsizes;
        }

        private byte[] GetContentHashes(byte[] tmdfile, int length)
        {
            /*string[] contenthashes = new string[length];
            int startoffset = 500;

            for (int i = 0; i < length; i++)
            {
                contenthashes[i] = MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 1]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 2]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 3]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 4]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 5]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 6]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 7]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 8]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 9]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 10]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 11]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 12]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 13]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 14]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 15]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 16]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 17]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 18]))) +
                    MakeProperLength(ConvertToHex(Convert.ToString(tmdfile[startoffset + 19])));
                //contentsizes[i] = TrimLeadingZeros(contentsizes[i]);
                startoffset += 36;
            }

            return contenthashes;*/
            byte[] contenthashes = new byte[length*20];
            int startoffset = 500;

            for (int i = 0; i < length; i++)
            {
                for (int x = 0; x < 20; x++)
                {
                    contenthashes[(i * 20) + x] = tmdfile[startoffset + x];
                }
                startoffset += 36;
            }
            //WriteStatus(DisplayBytes(contenthashes));
            return contenthashes;
        }

        private byte[] GetContentIndices(byte[] tmdfile, int length)
        {
            byte[] contentindices = new byte[length];
            int startoffset = 0x1E9;

            for (int i = 0; i < length; i++)
            {
                contentindices[i] = tmdfile[startoffset];
                startoffset += 36;
            }

            return contentindices;
        }

        private void button3_Click(object sender, EventArgs e)
        {
            // Prevent mass deletion
            if ((titleidbox.Text == "") && (titleversion.Text == ""))
            {
                WriteStatus("Please enter SOME info...");
                return;
            }

            // Running Downloads in background so no form freezing
            NUSDownloader.RunWorkerAsync();
        }

        private void NUSDownloader_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            // Preparations for Downloading
            Control.CheckForIllegalCrossThreadCalls = false;
            WriteStatus("Starting NUS Download. Please be patient!");
            button3.Enabled = false;
            titleidbox.Enabled = false;
            titleversion.Enabled = false;
            button3.Text = "Starting NUS Download!";

            // Current directory...
            string currentdir = Application.StartupPath;
            if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";

            // Prevent crossthread issues
            string titleid = titleidbox.Text;

            // Creates the directory 
            CreateTitleDirectory();

            // Wii / DSi
            bool wiimode = radioButton1.Checked;
            
            // Set UserAgent to Wii value
            generalWC.Headers.Add("User-Agent", "Opera/9.30 (Nintendo Wii; U; ; 2071; Wii Shop Channel/16.0(A); en)");

            // Get placement directory early...
            string titledirectory;
            if (titleversion.Text == "")
                titledirectory = currentdir + titleid + @"\";
            else
                titledirectory = currentdir + titleid + "v" + titleversion.Text + @"\";

            button3.Text = "Prerequisites: (0/2)";

            // Download TMD before the rest...
            string tmdfull = "tmd";
            if (titleversion.Text != "")
                tmdfull += "." + titleversion.Text;
            try
            {
                DownloadNUSFile(titleid, tmdfull, titledirectory, 0, wiimode);
            }
            catch (Exception)
            {
                WriteStatus("NUS (404): " + tmdfull);
                button3.Enabled = true;
                titleidbox.Enabled = true;
                titleversion.Enabled = true;
                button3.Text = "Start NUS Download!";
                dlprogress.Value = 0;
                DeleteTitleDirectory();
                return;
            }
            button3.Text = "Prerequisites: (1/2)";
            dlprogress.Value = 50;

            // Download CETK after tmd...
            try
            {
                DownloadNUSFile(titleid, "cetk", titledirectory, 0, wiimode);
            }
            catch (Exception)
            {
                if (ignoreticket.Checked == false)
                {
                    WriteStatus("NUS (404): cetk");
                    button3.Enabled = true;
                    titleidbox.Enabled = true;
                    titleversion.Enabled = true;
                    button3.Text = "Start NUS Download!";
                    dlprogress.Value = 0;
                    DeleteTitleDirectory();
                    return;
                }
                else
                {
                    WriteStatus("Ticket not found! Continuing, however WAD packing and decryption are not possible!");
                    packbox.Checked = false;
                    decryptbox.Checked = false;
                }
            }
            button3.Text = "Prerequisites: (2/2)";
            dlprogress.Value = 100;

            // Obtain TitleKey
            byte[] titlekey = new byte[16];
            if (decryptbox.Checked == true)
            { 
                // Create ticket file holder
                FileStream fs1 = File.OpenRead(titledirectory + @"\cetk");
                byte[] cetkbuf = ReadFully(fs1, 20);
                fs1.Close();

                // Load TitleKey into it's byte[]
                // It is currently encrypted...
                for (int i = 0; i < 16; i++)
			    {
                    titlekey[i] = cetkbuf[0x1BF + i];
			    }

                byte[] iv = new byte[16];
                for (int i = 0; i < 8; i++)
			    {
                    iv[i] = cetkbuf[0x1DC + i];
			    }
                for (int i = 0; i < 8; i++)
			    {
                    iv[i+8] = 0x00;
			    }

                // Common Key
                byte[] keyBytes;
                if (wiimode)
                    keyBytes = LoadCommonKey();
                else
                    keyBytes = LoadDSiCommonKey();

                initCrypt(iv, keyBytes);

                WriteStatus("Title Key: " + DisplayBytes(Decrypt(titlekey)));
                titlekey = Decrypt(titlekey);
            }

            // Read the tmd as a stream...
            FileStream fs = File.OpenRead(titledirectory + tmdfull);
            byte[] tmd = ReadFully(fs, 20);

            // Read Title Version...
            string tmdversion = "";
            for (int x = 476; x < 478; x++)
            {
                tmdversion += MakeProperLength(ConvertToHex(Convert.ToString(tmd[x])));
            }
            titleversion.Text = Convert.ToString(int.Parse(tmdversion, System.Globalization.NumberStyles.HexNumber));

            /*if ((currentdir + titleid + "v" + titleversion.Text + @"\") != titledirectory)
            {
                Directory.Move(titledirectory, currentdir + titleid + "v" + titleversion.Text + @"\");
                titledirectory = currentdir + titleid + "v" + titleversion.Text + @"\";
                DirectoryInfo di = new DirectoryInfo(titledirectory);
            } */

            // Read Content #...
            string contentstrnum = "";
            for (int x = 478; x < 480; x++)
            {
                contentstrnum += TrimLeadingZeros(Convert.ToString(tmd[x]));
            }
            WriteStatus("Content #: " + contentstrnum);
            button3.Text = "Content: (0/" + contentstrnum + ")";
            dlprogress.Value = 0;

            // Gather information...
            string[] tmdcontents = GetContentNames(tmd, Convert.ToInt32(contentstrnum));
            string[] tmdsizes = GetContentSizes(tmd, Convert.ToInt32(contentstrnum));
            byte[] tmdhashes = GetContentHashes(tmd, Convert.ToInt32(contentstrnum));
            byte[] tmdindices = GetContentIndices(tmd, Convert.ToInt32(contentstrnum));

            // Progress bar total size tally info...
            float totalcontentsize = 0;
            float currentcontentlocation = 0;
            for (int i = 0; i < tmdsizes.Length; i++)
            {
                totalcontentsize += int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber);
            }

            /* Show total size of download...
            int totaldlsize = 0;
            for (int i = 0; i < tmdsizes.Length; i++)
            {
                totaldlsize += Convert.ToInt32(tmdsizes[i]);
            }*/
            WriteStatus("Total Size: " + (long)totalcontentsize + " bytes");

            for (int i = 0; i < tmdcontents.Length; i++)
            {
                try
                {
                    if ((localuse.Checked) && (File.Exists(titledirectory + tmdcontents[i])))
                    {
                        WriteStatus("Leaving local " + tmdcontents[i] + ".");
                    }
                    else
                    {
                        DownloadNUSFile(titleid, tmdcontents[i], titledirectory, int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber), wiimode);
                    }
                }
                catch (Exception)
                {
                    WriteStatus("NUS (404): " + tmdcontents[i]);
                    button3.Enabled = true;
                    titleidbox.Enabled = true;
                    titleversion.Enabled = true;
                    button3.Text = "Start NUS Download!";
                    dlprogress.Value = 0;
                    DeleteTitleDirectory();
                    return;
                }

                button3.Text = "Content: (" + (i + 1) + @"/" + contentstrnum + ")";
                currentcontentlocation += int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber);

                if (decryptbox.Checked == true)
                {
                    // Create content file holder
                    FileStream cont = File.OpenRead(titledirectory + @"\" + tmdcontents[i]);
                    byte[] contbuf = ReadFully(cont, 20);
                    cont.Close();

                    // Create ticket file holder
                    FileStream fs1 = File.OpenRead(titledirectory + @"\cetk");
                    byte[] cetkbuf = ReadFully(fs1, 20);
                    fs1.Close();

                    byte[] iv = new byte[16];
                    for (int x = 0; x < 8; x++)
                    {
                        iv[x] = 0x00;
                    }
                    for (int x = 0; x < 7; x++)
                    {
                        iv[x + 8] = 0x00;
                    }
                    iv[1] = tmdindices[i];
                    

                    /*byte[] iv = new byte[16];
                    for (int x = 0; x < 8; x++)
                    {
                        iv[x] = cetkbuf[0x1DC + x];
                    }
                    for (int y = 0; y < 8; y++)
                    {
                        iv[y + 8] = 0x00;
                    }*/

                    initCrypt(iv, titlekey); 

                    // Create decrypted file
                    string zeros = "000000";
                    /*if (i >= 10)
                        zeros = "000000";
                    else
                        zeros = "0000000"; */

                    FileStream decfs = new FileStream(titledirectory + @"\" + zeros + i.ToString("X2") + ".app", FileMode.Create);
                    decfs.Write(Decrypt(contbuf), 0, int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber));
                    //decfs.Write(Decrypt(contbuf), 0, contbuf.Length);
                    decfs.Close();

                    WriteStatus("  - Decrypted: " + zeros + i.ToString("X2") + ".app");
                    //WriteStatus("  - IV: " + DisplayBytes(rijndaelCipher.IV));

                    byte[] hash = new byte[20];
                    for (int x = 0; x < 20; x++)
                    {
                        hash[x] = tmdhashes[(i * 20) + x];
                    }
                    byte[] deccont = Decrypt(contbuf);
                    Array.Resize(ref deccont, int.Parse(tmdsizes[i], System.Globalization.NumberStyles.HexNumber));
                    if ((Convert.ToBase64String(ComputeSHA(deccont))) == Convert.ToBase64String(hash))
                    {
                        WriteStatus("  - Hash Check: Pass");
                    }
                    else
                    {
                        WriteStatus("  - Hash Check: Fail");
                        WriteStatus("    - True Hash: " + DisplayBytes(hash));
                        WriteStatus("    - You Got: " + DisplayBytes(ComputeSHA(Decrypt(contbuf))));
                    }
                }

                dlprogress.Value = Convert.ToInt32(((currentcontentlocation / totalcontentsize) * 100));
            }

            WriteStatus("NUS Download Finished.");

            if ((packbox.Checked == true) && (wiimode == true))
            {
                PackWAD(titleid, tmdfull, tmdcontents.Length, tmdcontents, tmdsizes, titledirectory);
            }

            button3.Enabled = true;
            titleidbox.Enabled = true;
            titleversion.Enabled = true;
            button3.Text = "Start NUS Download!";
            dlprogress.Value = 0;

        }

        private void CreateTitleDirectory()
        {
            // Creates the directory for the downloaded title...
            string currentdir = Application.StartupPath;

            /*if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";
             */
            if (currentdir.EndsWith(Convert.ToString(Path.DirectorySeparatorChar)) == false)
                currentdir += Path.DirectorySeparatorChar;

            // Get placement directory early...
            string titledirectory;
            if (titleversion.Text == "")
                //titledirectory = currentdir + titleidbox.Text + @"\";
                titledirectory = Path.Combine(currentdir, titleidbox.Text + Path.DirectorySeparatorChar);
            else
                titledirectory = Path.Combine(currentdir, titleidbox.Text + "v" + titleversion.Text + Path.DirectorySeparatorChar);

            if ((localuse.Checked) && (Directory.Exists(titledirectory)))
            {
                //WriteStatus("Using Local Files");
            }
            else
            {
                if (Directory.Exists(titledirectory))
                    Directory.Delete(titledirectory, true);

                Directory.CreateDirectory(titledirectory);
            }
        }

        private void DeleteTitleDirectory()
        {
            string currentdir = Application.StartupPath;

            /*if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\"; */

            if (currentdir.EndsWith(Convert.ToString(Path.DirectorySeparatorChar)) == false)
                currentdir += Path.DirectorySeparatorChar;

            // Get placement directory early...
            string titledirectory;
            if (titleversion.Text == "")
                //titledirectory = currentdir + titleidbox.Text + @"\";
                titledirectory = Path.Combine(currentdir, titleidbox.Text + Path.DirectorySeparatorChar);
            else
                titledirectory = Path.Combine(currentdir, titleidbox.Text + "v" + titleversion.Text + Path.DirectorySeparatorChar);

            if (Directory.Exists(titledirectory))
                Directory.Delete(titledirectory, true);

            //Directory.CreateDirectory(currentdir + titleidbox.Text);
        }

        private void DownloadNUSFile(string titleid, string filename, string placementdir, int sizeinbytes, bool iswiititle)
        {
            string nusfileurl;
            if (iswiititle)
                nusfileurl = NUSURL + titleid + @"/" + filename;
            else
                nusfileurl = DSiNUSURL + titleid + @"/" + filename;

            WriteStatus("Grabbing " + filename + "...");

            if (sizeinbytes != 0)
                statusbox.Text += " (" + Convert.ToString(sizeinbytes) + " bytes)";

            generalWC.DownloadFile(nusfileurl, placementdir + filename);
        }

        public void PackWAD(string titleid, string tmdfilename, int contentcount, string[] contentnames, string[] contentsizes, string totaldirectory)
        {
            // Directory stuff
            string currentdir = Application.StartupPath;
            if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";

            // Create cert file holder
            FileStream certfs = File.OpenRead(currentdir + @"\cert.sys");
            byte[] certsbuf = ReadFully(certfs, 20);
            certfs.Close();

            // Create ticket file holder
            FileStream fs1 = File.OpenRead(totaldirectory + @"\cetk");
            byte[] cetkbuf = ReadFully(fs1, 20);
            fs1.Close();

            // Create tmd file holder
            FileStream fs2 = File.OpenRead(totaldirectory + @"\" + tmdfilename);
            byte[] tmdbuf = ReadFully(fs2, 20);
            fs2.Close();

            if (wadnamebox.Text.Contains("[v]") == true)
                wadnamebox.Text = wadnamebox.Text.Replace("[v]", "v" + titleversion.Text);

            // Create wad file
            FileStream wadfs = new FileStream(totaldirectory + @"\" + wadnamebox.Text, FileMode.Create);

            // Add wad stuffs
            WADHeader wad = new WADHeader();
            wad.HeaderSize = 0x20;
            wad.WadType = 0x49730000;
            wad.CertChainSize = 0xA00;

            // TMDSize is length of buffer.
            wad.TMDSize = tmdbuf.Length;
            // TicketSize is length of cetkbuf.
            wad.TicketSize = cetkbuf.Length;

            // Write cert[] to 0x40.
            wadfs.Seek(0x40, SeekOrigin.Begin);
            wadfs.Write(certsbuf, 0, certsbuf.Length);
            WriteStatus("Cert wrote (0x" + Convert.ToString(64, 16) + ")");

            // Need 64 byte boundary...
            wadfs.Seek(2624, SeekOrigin.Begin);

            // Cert is 2560
            // Write ticket at this point...
            wad.TicketSize = 0x2A4;
            wadfs.Write(cetkbuf, 0, wad.TicketSize);
            WriteStatus("Ticket wrote (0x" + Convert.ToString((wadfs.Length - 0x2A4), 16) + ")");

            // Need 64 byte boundary...
            wadfs.Seek(ByteBoundary(Convert.ToInt32(wadfs.Length)), SeekOrigin.Begin);

            // Write TMD at this point...
            wadfs.Write(tmdbuf, 0, 484 + (contentcount * 36));
            WriteStatus("TMD wrote (0x" + Convert.ToString((wadfs.Length - (484 + (contentcount * 36))), 16) + ")");

            // Preliminary data size of wad file.
            wad.DataSize = 0;

            // Loop n Add contents
            for (int i = 0; i < contentcount; i++)
            {
                // Need 64 byte boundary...
                wadfs.Seek(ByteBoundary(Convert.ToInt32(wadfs.Length)), SeekOrigin.Begin);

                // Create content file holder
                FileStream cont = File.OpenRead(totaldirectory + @"\" + contentnames[i]);
                byte[] contbuf = ReadFully(cont, 20);
                cont.Close();

                wadfs.Write(contbuf, 0, contbuf.Length);

                if (int.Parse(contentsizes[i], System.Globalization.NumberStyles.HexNumber) != contbuf.Length)
                {
                    // Content size mismatch
                    WriteStatus(contentnames[i] + " wrote (0x" + Convert.ToString((wadfs.Length - contbuf.Length), 16) + ") (Mismatch)");
                }
                else
                {
                    WriteStatus(contentnames[i] + " wrote (0x" + Convert.ToString((wadfs.Length - contbuf.Length), 16) + ")");
                }

                wad.DataSize += contbuf.Length;
            }

            wadfs.Seek(0, SeekOrigin.Begin);

            // Write initial part of header
            byte[] start = new byte[8] { 0x00, 0x00, 0x00, 0x20, 0x49, 0x73, 0x00, 0x00 };
            wadfs.Write(start, 0, start.Length);

            //WriteStatus("Initial header wrote");

            // Write CertChainLength
            wadfs.Seek(0x08, SeekOrigin.Begin);
            byte[] chainsize = InttoByteArray(wad.CertChainSize);
            wadfs.Write(chainsize, 0, 4);

            // Write res
            byte[] reserved = new byte[4] { 0x00, 0x00, 0x00, 0x00 };
            wadfs.Seek(0x0C, SeekOrigin.Begin);
            wadfs.Write(reserved, 0, 4);

            // Write ticketsize
            byte[] ticketsize = new byte[4] { 0x00, 0x00, 0x02, 0xA4 };
            wadfs.Seek(0x10, SeekOrigin.Begin);
            wadfs.Write(ticketsize, 0, 4);

            // Write tmdsize
            int strippedtmd = 484 + (contentcount * 36);
            byte[] tmdsize = InttoByteArray(strippedtmd);
            wadfs.Seek(0x14, SeekOrigin.Begin);
            wadfs.Write(tmdsize, 0, 4);

            // Write data size
            wadfs.Seek(0x18, SeekOrigin.Begin);
            byte[] datasize = InttoByteArray(wad.DataSize);
            wadfs.Write(datasize, 0, 4);

            WriteStatus("WAD Created: " + wadnamebox.Text);


            wadfs.Close();
        }

        private long ByteBoundary(int currentlength)
        {
            // Gets the next 0x40 offset.
            long thelength = currentlength - 1;
            long remainder = 1;

            while (remainder != 0)
            {
                thelength += 1;
                remainder = thelength % 0x40;
            }

            //WriteStatus("Initial Size: " + currentlength);
            //WriteStatus("0x40 Size: " + thelength);

            return (long)thelength;
        }

        private byte[] InttoByteArray(int size)
        {
            byte[] b = new byte[4];
            b = BitConverter.GetBytes(size);

            if (BitConverter.IsLittleEndian)
                Array.Reverse(b);

            return b;
        }

        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButton2.Checked == true)
            {
                // Cannot Pack WADs
                packbox.Checked = false;
                packbox.Enabled = false;

                // Can decrypt if key exists...lulz
                if (dsidecrypt == false)
                {
                    decryptbox.Checked = false;
                    decryptbox.Enabled = false;
                }

                wadnamebox.Enabled = false;
                wadnamebox.Text = "";
            }
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButton1.Checked == true)
            {
                // Can pack WADs
                // packbox.Checked = true;
                packbox.Enabled = true;
                decryptbox.Enabled = true;
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            // Display About Text...
            statusbox.Text = "";
            WriteStatus("NUS Downloader (NUSD)");
            WriteStatus("You are running version: " + version);
            WriteStatus("This program coded by WB3000");
            WriteStatus("");
            string currentdir = Application.StartupPath;
            if (currentdir.EndsWith(Convert.ToString(Path.DirectorySeparatorChar)) == false)
                currentdir += Path.DirectorySeparatorChar;
            if (File.Exists(currentdir + "key.bin") == false)
            {
                WriteStatus("Wii Decryption: Need (key.bin)");
            }
            else
            {
                WriteStatus("Wii Decryption: OK");
            }

            if (File.Exists(currentdir + "dskey.bin") == false)
            {
                WriteStatus("DSi Decryption: Need (dskey.bin)");
            }
            else
            {
                WriteStatus("DSi Decryption: OK");
            }
            
            WriteStatus("");
            WriteStatus("Special thanks to:");
            WriteStatus(" * Crediar for his wadmaker tool + source, and for the advice!");
            WriteStatus(" * SquidMan/Galaxy/comex for advice/sources.");
            WriteStatus(" * #WiiDev for general assistance whenever I had questions.");
        }

        private void getcerts_Click(object sender, EventArgs e)
        {
            // Directory stuff
            string currentdir = Application.StartupPath;
            if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";

            // Create certs file
            FileStream certsfs = new FileStream(currentdir + @"\cert.sys", FileMode.Create);

            DownloadNUSFile("0000000100000002", "tmd.289", currentdir + @"\", 0, true);
            DownloadNUSFile("0000000100000002", "cetk", currentdir + @"\", 0, true);

            // Create ticket file holder
            FileStream cetk = File.OpenRead(currentdir + "cetk");
            byte[] cetkbuf = ReadFully(cetk, 20);
            cetk.Close();

            // Create tmd file holder
            FileStream tmd = File.OpenRead(currentdir + "tmd.289");
            byte[] tmdbuf = ReadFully(tmd, 20);
            tmd.Close();

            certsfs.Seek(0, SeekOrigin.Begin);
            certsfs.Write(tmdbuf, 0x628, 0x400);
            WriteStatus("Added CA Cert!");

            certsfs.Seek(0x400, SeekOrigin.Begin);
            certsfs.Write(tmdbuf, 0x328, 0x300);
            WriteStatus("Added CACP Cert!");

            certsfs.Seek(0x700, SeekOrigin.Begin);
            certsfs.Write(cetkbuf, 0x2A4, 0x300);
            WriteStatus("Added CAXS Cert!");
            certsfs.Close();

            if (verifyMd5Hash(currentdir + @"\cert.sys", certs_MD5) == true)
            {
                WriteStatus("Certs File Successfully Created!");
            }
            else
            {
                WriteStatus("Error in Creating Certs File!");
                WriteStatus("Please report this error if you are sure it is not your internet connection");
            }

            foreach (Control ctrl in this.Controls)
            {
                ctrl.Enabled = true;
            }
            getcerts.Visible = false;
            wadnamebox.Enabled = false;
            File.Delete(currentdir + "cetk");
            File.Delete(currentdir + "tmd.289");

        }

        static string getMd5Hash(string input)
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            FileStream fs = new FileStream(input, FileMode.Open);
            MD5 md5 = new MD5CryptoServiceProvider();
            byte[] hash = md5.ComputeHash(fs);
            fs.Close();
            fs.Dispose();
            foreach (byte hex in hash)
            {
                //Returns hash in lower case.
                //To return upper case change “x2″ to “X2″
                sb.Append(hex.ToString("x2"));
            }
            return sb.ToString();
        }

        // Verify a hash against a string.
        static bool verifyMd5Hash(string input, string hash)
        {
            // Hash the input.
            string hashOfInput = getMd5Hash(input);

            // Create a StringComparer an comare the hashes.
            StringComparer comparer = StringComparer.OrdinalIgnoreCase;

            if (0 == comparer.Compare(hashOfInput, hash))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        private void packbox_CheckedChanged(object sender, EventArgs e)
        {
            if (packbox.Checked == true)
            {
                wadnamebox.Enabled = true;
                // Change WAD name if applicable
                if (packbox.Checked == true)
                {
                    if (titleidbox.Enabled == true)
                    {
                        if (titleversion.Text != "")
                        {
                            wadnamebox.Text = titleidbox.Text + "-NUS-v" + titleversion.Text + ".wad";
                        }
                        else
                        {
                            wadnamebox.Text = titleidbox.Text + "-NUS-[v]" + titleversion.Text + ".wad";
                        }
                    }
                }
            }
            else
            {
                wadnamebox.Enabled = false;
                wadnamebox.Text = "";
            }
        }

        private void titleidbox_TextChanged(object sender, EventArgs e)
        {
            // Change WAD name if applicable
            if (packbox.Checked == true)
            {
                if (titleidbox.Enabled == true)
                { 
                    if (titleversion.Text != "")
                    {
                        wadnamebox.Text = titleidbox.Text + "-NUS-v" + titleversion.Text + ".wad";
                    }
                    else
                    {
                        wadnamebox.Text = titleidbox.Text + "-NUS-[v]" + titleversion.Text + ".wad";
                    }
                }
            }
        }

        private void titleversion_TextChanged(object sender, EventArgs e)
        {
            // Change WAD name if applicable
            if (packbox.Checked == true)
            {
                if (titleidbox.Enabled == true)
                {
                    if (titleversion.Text != "")
                    {
                        wadnamebox.Text = titleidbox.Text + "-NUS-v" + titleversion.Text + ".wad";
                    }
                    else
                    {
                        wadnamebox.Text = titleidbox.Text + "-NUS-[v]" + titleversion.Text + ".wad";
                    }
                }
            }
        }

        public void initCrypt(byte[] iv, byte[] key)
        {
            rijndaelCipher = new RijndaelManaged();
            rijndaelCipher.Mode = CipherMode.CBC;
            rijndaelCipher.Padding = PaddingMode.None;
            rijndaelCipher.KeySize = 128;
            rijndaelCipher.BlockSize = 128;
            rijndaelCipher.Key = key;
            rijndaelCipher.IV = iv;
        }

        public byte[] Encrypt(byte[] plainBytes)
        {
            ICryptoTransform transform = rijndaelCipher.CreateEncryptor();
            using (MemoryStream ms = new MemoryStream(plainBytes))
            {
                using (CryptoStream cs = new CryptoStream(ms, transform, CryptoStreamMode.Read))
                {
                    return ReadFully(cs);
                }
            }
        }

        public byte[] Decrypt(byte[] encryptedData)
        {
            ICryptoTransform transform = rijndaelCipher.CreateDecryptor();
            using (MemoryStream ms = new MemoryStream(encryptedData))
            {
                using (CryptoStream cs = new CryptoStream(ms, transform, CryptoStreamMode.Read))
                {
                    return ReadFully(cs);
                }
            }
        }

        public byte[] ReadFully(Stream stream)
        {
            byte[] buffer = new byte[32768];
            using (MemoryStream ms = new MemoryStream())
            {
                while (true)
                {
                    int read = stream.Read(buffer, 0, buffer.Length);
                    if (read <= 0)
                        return ms.ToArray();
                    ms.Write(buffer, 0, read);
                }
            }
        }

        public string DisplayBytes(byte[] bytes)
        {
            string output = "";
            for (int i = 0; i < bytes.Length; ++i)
            {
                output += bytes[i].ToString("X2") + " ";
            }
            return output;
        }

        static public byte[] ComputeSHA(byte[] data)
        {
            SHA1 sha = new SHA1CryptoServiceProvider();
            // This is one implementation of the abstract class SHA1.
            return sha.ComputeHash(data);
        }

        public byte[] LoadCommonKey()
        {
            // Directory stuff
            string currentdir = Application.StartupPath;
            if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";

            if (File.Exists(currentdir + @"\key.bin") == true)
            {
                // Read common key into array
                FileStream ckey = File.OpenRead(currentdir + @"\key.bin");
                byte[] ckeybuf = ReadFully(ckey, 16);
                ckey.Close();
                return ckeybuf;
            }
            else
                return null;
        }

        public byte[] LoadDSiCommonKey()
        {
            // Directory stuff
            string currentdir = Application.StartupPath;
            if (!(currentdir.EndsWith(@"\")) || !(currentdir.EndsWith(@"/")))
                currentdir += @"\";

            if (File.Exists(currentdir + @"\dskey.bin") == true)
            {
                // Read common key into array
                FileStream ckey = File.OpenRead(currentdir + @"\dskey.bin");
                byte[] ckeybuf = ReadFully(ckey, 16);
                ckey.Close();
                return ckeybuf;
            }
            else
                return null;
        }
    }
}
